/***************************************************************************
              mp3block.cpp  -  implementation of MP3 sound block
             ----------------------------------------------------
    begin                : Sun Feb 13 2000
    copyright            : (C) 2000 by Ladislav Slezk
    email                : slezak@penguin.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "mp3block.h"
#include "sounddevice.h"
#include "utils.h"

#include <gnome--/app.h>
#include <libgnome/gnome-i18n.h>

#include <unistd.h>						// usleep + getcwd

#include <iostream.h>
#include <strstream.h>
#include <string.h>
#include <string>
#include <vector>

MP3block::MP3block() {
	scanning = false;
  exporting = false;
	audiobuffer_size = 1024;
	vbr = false;
	max_vbr = 0;
	tot_samples = -1;
	
  proc.stderr_data.connect(SigC::slot(this, &MP3block::received_err));
	proc.stdout_data.connect(SigC::slot(this, &MP3block::received_data));
	proc.process_exited.connect(SigC::slot(this, &MP3block::player_exited));
	proc.data_written.connect(SigC::slot(this, &MP3block::data_written));

}

MP3block::MP3block(const MP3block &src) : SoundBlock(src) {
	scanning = false;
  exporting = false;
	audiobuffer_size = 1024;
	
  proc.stderr_data.connect(SigC::slot(this, &MP3block::received_err));
	proc.stdout_data.connect(SigC::slot(this, &MP3block::received_data));
	proc.process_exited.connect(SigC::slot(this, &MP3block::player_exited));
	proc.data_written.connect(SigC::slot(this, &MP3block::data_written));
	
	// copy data members
	bitrate = src.bitrate;
	layer = src.layer;
	MPEGversion = src.MPEGversion;
	framesize = src.framesize;

	frameOffs = src.frameOffs;
	vbr = src.vbr;
	max_vbr = src.max_vbr;
	crc = src.crc;
	private_flag = src.private_flag;
	original = src.original;
	copyright = src.copyright;
	stereo_mode = src.stereo_mode;
	
	tot_samples = src.tot_samples;
}

MP3block::~MP3block() {
}

bool MP3block::init_play() {
#ifdef DEBUG_OUTPUT
  cerr << "Init play" << endl;
#endif

	pausebuffer = 0;
	sizeofpausebuffer = 0;
	usedBuffer = 0;

  return true;		
}

bool MP3block::play_block(SoundDevice *sd, string mpg123_path, int from_ms = -1, int to_ms = -1) {
#ifdef DEBUG_OUTPUT
  cerr << "MP3 Play block - decode with: " << mpg123_path << endl;
#endif

  playedMS = 0;
	play_from_offs_ms = 0;
	totalReceived = 0;
	usedBuffer = 0;
  sizeofpausebuffer = 0;

  play_from_ms = from_ms;
  play_to_ms = to_ms;

	if (!assigned) return false;
	playing = true;
	
  sd->close_device();		// reopen sound device - required befor setting new frequency, format, ...
  sd->open_device();

  // set sound format according to machine architecture
#ifdef WORDS_BIGENDIAN
  sd->set_format(SoundDevice::bit16_be);		// output is 16 bit, big endian byte order
#else
  sd->set_format(SoundDevice::bit16_le);		// output is 16 bit, little endian byte order
#endif

	sd->set_number_of_channels(channels);							// set channels
	sd->set_frequency(freq);										// set frequency
	
	sdev = sd;      // sound device
	
	audiobuffer = new char[audiobuffer_size];	

	if (!file.is_open()) file.open(fileName.c_str(), ios::binary | ios::in);			// open file

  int start = begin;

  if (from_ms > 0 && from_ms < block_length()) {
#ifdef DEBUG_OUTPUT
    cerr << "MP3 Play block play from" << endl;
#endif

    int frOffset = static_cast<int>(play_from_ms / 1000.0 * freq) / samples_per_frame();
    play_from_offs_ms = frOffset * samples_per_frame() / freq * 1000;
    int index = 0, sum = 0;

    while (sum < begin) {
      sum += frameOffs[index++];
    }

    start = sum;

    while (frOffset > 0) {
      start += frameOffs[index++];
      frOffset--;
    }
  }

  if (to_ms > 0 && to_ms < block_length()) {
#ifdef DEBUG_OUTPUT
    cerr << "MP3 Play block play to" << endl;
#endif

    int frOffset = static_cast<int>(play_to_ms / 1000.0 * freq) / samples_per_frame();
    int index = 0, sum = 0;

    while (sum < begin) {
      sum += frameOffs[index++];
    }

    finish = sum;

    while (frOffset > 0) {
      finish += frameOffs[index++];
      frOffset--;
    }
  }
  else
    finish = end;

  if (start >= finish) {
  	delete[] audiobuffer;
  	audiobuffer = 0;
  	playing = false;
    block_done.emit();
    return false;
  }

  file.seekg(start, ios::beg);

  vector<string> s;
  s.push_back("-s");
  s.push_back("-");
  proc.start(mpg123_path, s);

	data_written();
	
  return true;
}

bool MP3block::is_changed() {
  return changed;
}

bool MP3block::scan_file(int b, int e) {	// first sample, last sample
	scanning = true;
	totalReceived = 0;
	usedBuffer = 0;
	audiobuffer = new char[audiobuffer_size];	
	pausebuffer = 0;
	sizeofpausebuffer = 0;
	play_to_ms = -1;
	
	num = 2000;												// number of scan points

  if (e < b) {
    int tmp = b;
  	b = e;
  	e = tmp;
  }

  beginScan = b;
  endScan = e;

  if (e - b < num) num = e - b;

  progress = manage(new DialogProgress(_("Please wait ..."), _("Scanning MP3 file ...")));
  progress->canceled.connect(slot(this, &MP3block::scanning_canceled));

  rat = static_cast<double>(e - b) / num;

  scannedData.erase(scannedData.begin(), scannedData.end());    // clear scanned data


  switch (freq) {                   // set maximum downsample
  	case(44100) :
  	case(48000)	: downSample = 4; break;

  	case(22050) :
  	case(24000)	: downSample = 2; break;
  	
  	default : downSample = 1; break;
  }


  strstream ss;
  ss << downSample;

  string s;
  ss >> s;

#ifdef DEBUG_OUTPUT
  cerr << "Downsample: " << downSample << endl;
#endif
		
  file.seekg(begin, ios::beg);

  vector<string> pars;
  pars.push_back("-s");
  pars.push_back("-" + s);
  pars.push_back("-");
  proc.start("mpg123", pars);

	data_written();

  return true;
}

int MP3block::block_length() {	// number of samples in block
	return static_cast<int>(static_cast<double>(num_samples()) / freq * 1e3);
}

bool MP3block::valid() {			// is this soundblock valid?
  return true;
}

string MP3block::block_info() {	
  strstream ss;
  string s;

  ss << bitrate;
  ss >> s;

  string result;

  if (!vbr) {
    result = s + _(" kb/s");
  }
  else {
    strstream ss2;
    string s2;

    ss2 << max_vbr;
    ss2 >> s2;

    result = "VBR " + s + "-" + s2 + _(" kb/s");
  }

  string mode;
  switch(stereo_mode) {
    case (MODE_STEREO): mode = _("stereo"); break;
    case (MODE_JOINT_STEREO): mode = _("joint stereo"); break;
    case (MODE_DUAL): mode = _("dual stereo"); break;
    case (MODE_MONO): mode = _("mono"); break;
  }

  string crc_str;
  if (crc) crc_str = ", CRC";

  return result + ", " + mode + crc_str;
}

bool MP3block::assign_file(string fname) {
  char hdr[4];

  file.open(fname.c_str(), ios::binary | ios::in);			// open file

  if (!file.is_open())
  	return false;  						
  		
  fileName = fname;

  // test if ID3v2 tag is present
  file.read(hdr, 3);
  hdr[3] = '\0';

  if (strcmp(hdr, "ID3") == 0) {
#ifdef DEBUG_OUTPUT
    cerr << "ID3v2 Tag present" << endl;
#endif

    // tag present
    file.seekg(5, ios::beg);   // skip to tag length
    file.read(hdr, 1);
    bool unsync = (static_cast<unsigned char>(hdr[0]) > static_cast<unsigned char>(127));

#ifdef DEBUG_OUTPUT
    if (unsync) cerr << "Unsynchronization is used" << endl;
#endif

    file.read(hdr, 1);

    int taglen = static_cast<int>(hdr[0]);

    file.read(hdr, 1);
    taglen = taglen << 7;
    taglen = taglen | static_cast<int>(hdr[0]);

    file.read(hdr, 1);
    taglen = taglen << 7;
    taglen = taglen | static_cast<int>(hdr[0]);

    file.read(hdr, 1);
    taglen = taglen << 7;
    taglen = taglen | static_cast<int>(hdr[0]);

#ifdef DEBUG_OUTPUT
    cerr << "Tag length: " << taglen << endl;
#endif

    begin = 10 + taglen;    // 10 is size of tag header
    file.seekg(begin, ios::beg);
  }
  else {
    file.seekg(0, ios::beg);       // tag not present, set position to beginning
    begin = 0;				// begin of sound data
  }

  guint32 head;
  file.read(&head, sizeof(head));    // read header from file

#ifndef WORDS_BIGENDIAN
  head = GUINT32_FROM_BE(head);
#endif


#ifdef DEBUG_OUTPUT
  cerr << "Head: " << head << endl;
#endif

	if (!hdr_sync(head)) return false;

	// read values from header
	freq = hdr_frequency(head);
	channels = hdr_channels(head);
	bitrate = hdr_bitrate(head);
	framesize = hdr_frame_size(head);
	layer = hdr_layer(head);
	MPEGversion = hdr_mpeg_version(head);
	
	crc = hdr_protection(head);
	private_flag = hdr_private(head);
	original = hdr_original(head);
	copyright = hdr_copyright(head);
	stereo_mode = hdr_channel_mode(head);
	
#ifdef DEBUG_OUTPUT
	cerr << "crc       : " << crc << endl;
	cerr << "original  : " << original << endl;
	cerr << "private   : " << private_flag << endl;
	cerr << "copyright : " << copyright << endl;
#endif

	file.seekg(0, ios::end);        // get file size	
	int file_size = file.tellg();
	
#ifdef DEBUG_OUTPUT
	cerr << "File size: " << file_size << endl;
#endif
	
  if (!has_id3v1_tag(file)) {			// set end of sound data
		end = file_size;
	}
	else {
		end = file_size - 128;  // subtract size of ID3 tag
#ifdef DEBUG_OUTPUT
    cerr << "ID3v1 Tag present" << endl;
#endif
  }		

  beginMax = begin;
  endMax = end;
  assigned = true;

#ifdef DEBUG_OUTPUT
  cerr << "begin:" << begin << "  end: " << end << endl;
#endif

  scanOffs();     // get frame offsets

  // get time of last modification - write here

  return true;
}

bool MP3block::assigned_file() {
	return assigned;
}

int MP3block::num_samples() {          // number of frames * number of samples in frame
  if (frameOffs.size() == 0)
  	return (end - begin) / framesize * samples_per_frame();    // no offsets - estimate numsamples
  else {
    if (tot_samples < 0) tot_samples = (get_end_frame() - get_begin_frame()/* + 1*/) * samples_per_frame();
    return tot_samples;
//      return (get_end_frame() - get_begin_frame()) * samples_per_frame();
  }
}

int MP3block::max_samples() {          // number of frames * number of samples in frame
  if (frameOffs.size() == 0)
  	return (endMax - beginMax) / framesize * samples_per_frame();     // no offsets - estimate numsamples
  else
    return frameOffs.size() * samples_per_frame();
}


int MP3block::read_sample(int pos, int channel) {
  return 0;
}

void MP3block::stop_block() {
  if (playing) {
  	playing = false;
  	if (!proc.kill_process()) proc.kill_process(SIGKILL);   // terminate (or kill) decoder
		sdev->reset();      // immediately stop playing
		sdev->close_device();      // close device
///!!		delete[] pausebuffer;
  }
}

void MP3block::pause_block() {
  if (playing) {
    paused = true;
  }
}

void MP3block::unpause_block() {
  if (playing) {
  	paused = false;
  	data_written();
  }
}

string MP3block::get_type(){
	string l, v;
	
  strstream ss;
  ss << layer;
  ss >> l;
	
	switch (MPEGversion) {
	  case (1)  : v = "1.0"; break;
	  case (2)  : v = "2.0"; break;
	  case (25) : v = "2.5"; break;
	
	  default   : v = ""; break;
	}
	
	return "MPEG " + v + " layer " + l;
}

bool MP3block::hdr_sync(guint32 h) {
	return ((h & 0xffe00000) == 0xffe00000);
}

int MP3block::hdr_mpeg_version(guint32 h) {
	unsigned int tmp = (h >> 19);
	tmp = tmp & 0x3;
	
	switch(tmp){
		case(0) : return 25;			// MPEG 2.5
		case(1) : return 0;   		// reserved
		case(2) : return 2;       // MPEG 2
		case(3) : return 1;       // MPEG 1
		default : return 0;	
	}
}

int MP3block::hdr_layer(guint32 h) {
	unsigned int tmp = (h >> 17) & 0x3;
	
	switch(tmp){
		case(0) : return 0;			// reserved
		case(1) : return 3;   	// Layer 3
		case(2) : return 2;     // Layer 2
		case(3) : return 1;     // Layer 1
		default : return 0;	
	}
}

int MP3block::hdr_protection(guint32 h) {
	int tmp = h >> 16;
	return !(tmp & 0x1);
}

int MP3block::hdr_bitrate(guint32 h) {
	int tmp = (h >> 12) & 0xf;

	int mpeg1layer1[] = {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, 0};
	int mpeg1layer2[] = {0, 32, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, 0};
	int mpeg1layer3[] = {0, 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 0};
	
	int mpeg2layer1[] = {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256, 0};
	int mpeg2layer23[] = {0, 8, 16, 24, 32, 40, 48, 56, 64, 80, 96, 112, 128, 144, 160, 0};

	switch(hdr_mpeg_version(h)) {
		case(1) : {
			switch(hdr_layer(h)) {
				case(1) : return mpeg1layer1[tmp];
				case(2) : return mpeg1layer2[tmp];
				case(3) : return mpeg1layer3[tmp];
			  default : return 0;
			}
		}
		
		case(2) :
		case(3) :
				switch(hdr_layer(h)) {
				case(1) : return mpeg2layer1[tmp];
				
				case(2) :
				case(3) : return mpeg2layer23[tmp];
			  default : return 0;
			}
	
		default : return 0;	
	}
	
}

int MP3block::hdr_frequency(guint32 h) {
	int tmp = (h >> 10) & 0x3;
	int ver = hdr_mpeg_version(h);

	if (ver == 1) {				// Mpeg 1
  	switch(tmp) {
  		case(0) : return 44100;
  		case(1) : return 48000;
  		case(2) : return 32000;
  		case(3) : return 0;
  		
  		default : return 0;
  	}
	}

	if (ver == 2) {				// Mpeg 2
  	switch(tmp) {
  		case(0) : return 22050;
  		case(1) : return 24000;
  		case(2) : return 16000;
  		case(3) : return 0;
  		
  		default : return 0;
  	}
	}
		
	if (ver == 25) {				// Mpeg 2.5
  	switch(tmp) {
  		case(0) : return 11025;
  		case(1) : return 12000;
  		case(2) : return 8000;
  		case(3) : return 0;
  		
  		default : return 0;
  	}
	}
	
	return 0;		// unknown MPEG version
}

int MP3block::hdr_padding(guint32 h) {
	int tmp = (h >> 9) & 0x1;
	
	if (tmp == 0) return 0;
	
	if (hdr_layer(h) == 1) return 4;
	else return 1;
}

int MP3block::hdr_private(guint32 h) {
	int tmp = h >> 8;
	return tmp & 0x1;
}

int MP3block::hdr_channels(guint32 h) {
	int tmp = (h >> 6) & 0x3;
	
  if (tmp == 3) return 1;				// mono
  else return 2;								// stereo or dual
}

MP3block::ChannelMode MP3block::hdr_channel_mode(guint32 h) {
	int tmp = (h >> 6) & 0x3;
	
	switch (tmp) {
	  case(0): return MODE_STEREO;
	  case(1): return MODE_JOINT_STEREO;
	  case(2): return MODE_DUAL;
	  case(3): return MODE_MONO;
	  default: return MODE_JOINT_STEREO;    // suppress warning
	}
}


int MP3block::hdr_mode_extension(guint32 h) {
	return (h >> 4) & 0x3;

}

int MP3block::hdr_copyright(guint32 h) {
	int tmp = h >> 3;
	return tmp & 0x1;
}

int MP3block::hdr_original(guint32 h) {
	int tmp = h >> 2;
	return tmp & 0x1;
}

int MP3block::hdr_emphasis(guint32 h) {
	return h & 0x3;
}

int MP3block::hdr_frame_size(guint32 h) {
	if (hdr_frequency(h) == 0) return 0;

	if (hdr_mpeg_version(h) == 1) {
  	switch(hdr_layer(h)) {         // MPEG 1.0
  		case(1) : return 12000 * hdr_bitrate(h) / hdr_frequency(h) + hdr_padding(h);
  	
  		case(2) :
  		case(3) : return 144000 * hdr_bitrate(h) / hdr_frequency(h) + hdr_padding(h);
  		
  		default : return 0;
  	}
  }
  else {    // MPEG 2.0     (how get frame length MPEG 2.5???)
  	switch(hdr_layer(h)) {
  		case(1) : return 24000 * hdr_bitrate(h) / hdr_frequency(h) + hdr_padding(h);
  	
  		case(2) :
  		case(3) : return 72000 * hdr_bitrate(h) / hdr_frequency(h) + hdr_padding(h);
  		
  		default : return 0;
  	}
  }
}

bool MP3block::has_id3v1_tag(ifstream &f) {
  char hdr[4];
	int pos = f.tellg();

  // test if ID3v1 tag is present
  f.seekg(-128, ios::end);
  f.read(hdr, 3);
  hdr[3] = '\0';

  f.seekg(pos, ios::beg);   // set previous position
	
	if (strcmp(hdr, "TAG") == 0)
	  return true;
	else
	  return false;
}

void MP3block::player_exited(int ret) {
#ifdef DEBUG_OUTPUT
	cerr << "MP3 block: mpg123 exited:" << ret << endl;
#endif
	
	if (playing) sdev->close_device();
	delete[] audiobuffer;
	audiobuffer = 0;
	
	if (scanning) {
		scanning = false;
#ifdef DEBUG_OUTPUT
  	cerr << "MP3 block: scan done" << endl;
#endif

    progress->destroy();

		scan_done.emit();
		return;
	}
	
	if (exporting) {
	  exporting = false;
#ifdef DEBUG_OUTPUT
  	cerr << "MP3 block: export done" << endl;
#endif

    export_done.emit();
    return;
	}
	
	if (playing) {
  	playing = false;
    block_done.emit();
    return;
  }
}

void MP3block::data_written() {

	if (paused)	return;
	
	if (!exporting && usedBuffer > 0) {				// play paused buffer
		int bsize = sdev->get_fragment_size();
    int num = (usedBuffer) / bsize;
    int rem = (usedBuffer) % bsize;

    int i = 0;
    while (i < num)	{
			sdev->play(pausebuffer + i*bsize, bsize);
      totalReceived += bsize;
//   	  playedMS = static_cast<int>(static_cast<double>(totalReceived) / channels / 2 / freq * 1000) + play_from_offs_ms;
//  		playedMiliSec.emit(playedMS);	// emit played micro secs
			i++;
    }

    if (rem > 0) {
			sdev->play(pausebuffer + num*bsize, rem);
      totalReceived += rem;
//   	  playedMS = static_cast<int>(static_cast<double>(totalReceived) / channels / 2 / freq * 1000) + play_from_offs_ms;
//  		playedMiliSec.emit(playedMS);	// emit played micro secs
    }

    usedBuffer = 0;
	}
	
	int posF = file.tellg();
	int tmpend;
	
	if (play_to_ms > 0)
	  tmpend = finish;
	else
	  tmpend = end;
	
	if (posF < tmpend) {
		if (tmpend - posF >= audiobuffer_size) {
    	file.read(audiobuffer, audiobuffer_size);
    	proc.write_stdin(audiobuffer, audiobuffer_size);
  	}
  	else {
  		int readLen = tmpend - posF;
    	file.read(audiobuffer, readLen);
    	proc.write_stdin(audiobuffer, readLen);
  	}
	}
	else
		proc.close_stdin();
	
}


void MP3block::received_data(char *data, int len) {
	if (scanning) {
		int numPoints = scannedData.size() / channels;
		int maxPoints = static_cast<int>((totalReceived + len) / channels / 2 / rat * downSample);
		
		while (numPoints < maxPoints) {
			int index = len - (totalReceived + len) / channels / 2 % static_cast<int>(rat);
			int err = (totalReceived + index) % (2 * channels);
			
      if (numPoints % 10 == 0) {
  	    progress->set_progress(static_cast<float>(file.tellg()) / end * 100);
	    }
			
			index = abs(index);
			
			if (err != 0) {
				index -= err;
				if (index < 0) index += 4;
			}
			
			if (index > len - 4) index = len - 4;
			
			char x, y;
			x = *(data + index + 1);
			y = *(data + index);
			gint16 value = (x << 8) + y;
			
			int snum = 0;
			gint16 value2;
			
			while(snum <= 128 && (index + snum + 1) < len) {
				x = *(data + snum + index + 1);
				y = *(data + snum + index);
				value2 = (x << 8) + y;
				
				if (value2 > value) value = value2;
				snum += 4;
				
			}
			scannedData.push_back(value);
			
			if (channels == 2) {
  			x = *(data + index + 1 + 2);
  			y = *(data + index + 2);
  			gint16 value = (x << 8) + y;
  			
  			snum = 0;
   			
				while(snum <= 128 && (index + snum + 1) < len) {
  				x = *(data + snum + index + 1 + 2);
  				y = *(data + snum + index + 2);
  				value2 = (x << 8) + y;
  				
  				if (value2 > value) value = value2;
  				snum += 4;
				}
   			   			
				scannedData.push_back(value);
			}
			
		  numPoints++;
		}
			
    totalReceived += len;

    return;
	}
	
	if (exporting) {
#ifdef WORDS_BIGENDIAN
     // convert 16 bit data on big endian system to little endian - WAV is little endian
     if (exporting_WAV && bitsPerSample == 16) {			
       gint16 *conv = reinterpret_cast<gint16 *>(data);
       int lng = len / 2;
       for (int i = 0; i < lng; i++) {
         conv[i] = GINT16_TO_LE(conv[i]);
       }
     }
#endif

    export_stream->write(data, len);

	  if (block_num % 20 == 0) {
      float percent = static_cast<float>(file.tellg() - begin) / (end - begin) * 100.0f;
      export_progress->set_progress(percent);
//   	  while (Gtk::Main::events_pending() > 0) Gtk::Main::iteration();
   	
  	  if (export_progress->cancel_state()) {
	      proc.kill_process();
	      return;
	    }
 	  }

    block_num++;
	  return;
	}

	if (!paused) {
  	sdev->play(data, len);
    totalReceived += len;
// 	  playedMS = static_cast<int>(static_cast<double>(totalReceived) / channels / 2 / freq * 1000) + play_from_offs_ms;;
//  	playedMiliSec.emit(playedMS);	// emit played micro secs
	}
	else {
	
#ifdef DEBUG_OUTPUT
	  cerr << "Mp3block: paused  received: " << len << endl;
#endif

		if (sizeofpausebuffer == 0) {					// create buffer 32kB
			sizeofpausebuffer = 32 * 1024;
      pausebuffer = new char[sizeofpausebuffer];
      memcpy(pausebuffer, data, len);
      usedBuffer = len;
    }
    else {
    	if (usedBuffer + len <= sizeofpausebuffer) {        // append to buffer
      	memcpy(pausebuffer + usedBuffer, data, len);
      	usedBuffer += len;
     	}
    	else {
      	char *buff = new char[sizeofpausebuffer + 32 * 1024];		// reallocate buffer + 32kB
        memcpy(buff, pausebuffer, usedBuffer);
        memcpy(buff + usedBuffer, data, len);

        delete[] pausebuffer;

        pausebuffer = buff;
        usedBuffer += len;
        sizeofpausebuffer += 32 * 1024;
			}
    }
	}
}

void MP3block::received_err(char *d, int len) {
#ifdef DEBUG_OUTPUT
	int i = 0;
  while(i < len) {
		cerr << d[i++];
  }
  cerr << endl;
#endif
}

bool MP3block::save_block_to_stream(ofstream &of, string dirName) {
	of << 'M';			// MPEG identification
	
	string relName;
	
	if (fileName.size() > 0 && fileName[0] != '/')
  	relName = fileName;
  else
  	relName = Utils::relativeFile(fileName, dirName);
	
	
  of.write(relName.c_str(), relName.size());       // save relative file to project directory
  of << '\0';
//  of.write((char *)(&fileDateTime), sizeof(fileDateTime));

  gint32 tmp32;
  gint16 tmp16;

#ifdef WORDS_BIGENDIAN
  tmp32 = GINT32_TO_LE(freq);
  of.write((char *)(&tmp32), sizeof(tmp32));
#else
  of.write((char *)(&freq), sizeof(freq));
#endif

#ifdef WORDS_BIGENDIAN
  tmp16 = GINT16_TO_LE(channels);
  of.write((char *)(&tmp16), sizeof(tmp16));
#else
  of.write((char *)(&channels), sizeof(channels));
#endif

#ifdef WORDS_BIGENDIAN
  tmp32 = GINT32_TO_LE(begin);
  of.write((char *)(&tmp32), sizeof(tmp32));
#else
  of.write((char *)(&begin), sizeof(begin));
#endif

#ifdef WORDS_BIGENDIAN
  tmp32 = GINT32_TO_LE(end);
  of.write((char *)(&tmp32), sizeof(tmp32));
#else
  of.write((char *)(&end), sizeof(end));
#endif

#ifdef WORDS_BIGENDIAN
  tmp32 = GINT32_TO_LE(beginMax);
  of.write((char *)(&tmp32), sizeof(tmp32));
#else
  of.write((char *)(&beginMax), sizeof(beginMax));
#endif

#ifdef WORDS_BIGENDIAN
  tmp32 = GINT32_TO_LE(endMax);
  of.write((char *)(&tmp32), sizeof(tmp32));
#else
  of.write((char *)(&endMax), sizeof(endMax));
#endif


	// MPEG specific
#ifdef WORDS_BIGENDIAN
  tmp16 = GINT16_TO_LE(bitrate);
  of.write((char *)(&tmp16), sizeof(tmp16));
#else
  of.write((char *)(&bitrate), sizeof(bitrate));
#endif

#ifdef WORDS_BIGENDIAN
  tmp16 = GINT16_TO_LE(layer);
  of.write((char *)(&tmp16), sizeof(tmp16));
#else
  of.write((char *)(&layer), sizeof(layer));
#endif

#ifdef WORDS_BIGENDIAN
  tmp16 = GINT16_TO_LE(MPEGversion);
  of.write((char *)(&tmp16), sizeof(tmp16));
#else
  of.write((char *)(&MPEGversion), sizeof(MPEGversion));
#endif

#ifdef WORDS_BIGENDIAN
  tmp32 = GINT32_TO_LE(framesize);
  of.write((char *)(&tmp32), sizeof(tmp32));
#else
  of.write((char *)(&framesize), sizeof(framesize));
#endif

  of.write((char *)(&vbr), sizeof(vbr));

#ifdef WORDS_BIGENDIAN
  tmp16 = GINT16_TO_LE(max_vbr);
  of.write((char *)(&tmp16), sizeof(tmp16));
#else
  of.write((char *)(&max_vbr), sizeof(max_vbr));
#endif

  of.write((char *)(&crc), sizeof(crc));
  of.write((char *)(&private_flag), sizeof(private_flag));
  of.write((char *)(&original), sizeof(original));
  of.write((char *)(&copyright), sizeof(copyright));

#ifdef WORDS_BIGENDIAN
  tmp16 = GINT16_TO_LE(static_cast<gint16>(stereo_mode));
  of.write((char *)(&tmp16), sizeof(tmp16));
#else
  of.write((char *)(&stereo_mode), sizeof(stereo_mode));
#endif

	if (saveScannedData) {
    of << '\1';

    gint32 s = scannedData.size();
#ifdef WORDS_BIGENDIAN
    tmp32 = GINT32_TO_LE(s);
    of.write((char *)(&tmp32), sizeof(tmp32));
#else
    of.write((char *)(&s), sizeof(s));
#endif

    int i = 0;
    while (i < s) {
#ifdef WORDS_BIGENDIAN
      tmp16 = GINT16_TO_LE(scannedData[i]);
      of.write((char *)(&tmp16), sizeof(tmp16));
#else
      of.write((char *)(&scannedData[i]), sizeof(scannedData[i]));
#endif
      i++;
    }
	}
	else {
    of << '\0';
	}

		
	// save frame offsets if present
	if (frameOffs.size() > 0) {
    of << '\1';

    gint32 s = frameOffs.size();
#ifdef WORDS_BIGENDIAN
      tmp32 = GINT32_TO_LE(s);
      of.write((char *)(&tmp32), sizeof(tmp32));
#else
    of.write((char *)(&s), sizeof(s));
#endif

    int i = 0;
    while (i < s) {
#ifdef WORDS_BIGENDIAN
      tmp32 = GINT32_TO_LE(frameOffs[i]);
      of.write((char *)(&tmp32), sizeof(tmp32));
#else
      of.write((char *)(&frameOffs[i]), sizeof(frameOffs[i]));
#endif

      i++;
    }
	}
	else {
    of << '\0';
	}

	return true;
}

bool MP3block::load_block_from_stream(ifstream &is) {
  string q;
  char x;

  is.read(&x, sizeof(x));

  while (x != '\0') {
    q = q + x;
    is.read(&x, sizeof(x));
  }

  // convert file name from relative to absolute
  char *cdir = g_get_current_dir();
  fileName = Utils::absoluteFile(q, cdir);
  g_free(cdir);

#ifdef DEBUG_OUTPUT
  cerr << fileName << endl;
#endif

  file.open(fileName.c_str(), ios::binary | ios::in);			// open file

  if (!file.is_open())
  	return false;  						
  		
  assigned = true;

//  is.read(&fileDateTime, sizeof(fileDateTime));
  is.read(&freq, sizeof(freq));
  is.read(&channels, sizeof(channels));
  is.read(&begin, sizeof(begin));
  is.read(&end, sizeof(end));
  is.read(&beginMax, sizeof(beginMax));
  is.read(&endMax, sizeof(endMax));

#ifdef WORDS_BIGENDIAN
  // convert from little endian
  freq = GINT32_FROM_LE(freq);
  channels = GINT16_FROM_LE(channels);
  begin = GINT32_FROM_LE(begin);
  end = GINT32_FROM_LE(end);
  beginMax = GINT32_FROM_LE(beginMax);
  endMax = GINT32_FROM_LE(endMax);
#endif

#ifdef DEBUG_OUTPUT
  cerr << "Freq:  " << freq << endl;
  cerr << "Chan:  " << channels << endl;
  cerr << "Begin:  " << begin << endl;
  cerr << "End:  " << end << endl;
#endif

	// MPEG specific
  is.read(&bitrate, sizeof(bitrate));
  is.read(&layer, sizeof(layer));
  is.read(&MPEGversion, sizeof(MPEGversion));
  is.read(&framesize, sizeof(framesize));
  is.read(&vbr, sizeof(vbr));
  is.read(&max_vbr, sizeof(max_vbr));
  is.read(&crc, sizeof(crc));
  is.read(&private_flag, sizeof(private_flag));
  is.read(&original, sizeof(original));
  is.read(&copyright, sizeof(copyright));

  gint tmp16;
  is.read(&tmp16, sizeof(tmp16));

#ifdef WORDS_BIGENDIAN
  // convert from little endian
  bitrate = GINT16_FROM_LE(bitrate);
  layer = GINT16_FROM_LE(layer);
  MPEGversion = GINT16_FROM_LE(MPEGversion);
  framesize = GINT32_FROM_LE(framesize);
  max_vbr = GINT16_FROM_LE(max_vbr);
  tmp16 = GINT16_FROM_LE(tmp16);
#endif

  stereo_mode = static_cast<ChannelMode>(tmp16);


  // load scanned data
  is.read(&x, sizeof(x));

	if (x == '\1') {             // scanned data present
    gint32 size;               // number of scanned data

		is.read(&size, sizeof(size));
#ifdef WORDS_BIGENDIAN
    // convert from little endian
    size = GINT32_FROM_LE(size);
#endif

    int i = 0;
    gint16 i16;

    while (i < size) {
		  is.read(&i16, sizeof(i16));
#ifdef WORDS_BIGENDIAN
      // convert from little endian
      i16 = GINT16_FROM_LE(i16);
#endif
		
      scannedData.push_back(i16);
      i++;
    }
	}
	
  // load frame offsets
  is.read(&x, sizeof(x));

	if (x == '\1') {             // offsets present
    gint32 size;               // number of offsets
		is.read(&size, sizeof(size));
#ifdef WORDS_BIGENDIAN
    // convert from little endian
    size = GINT32_FROM_LE(size);
#endif

    int i = 0;
    gint32 i32;

    while (i < size) {
		  is.read(&i32, sizeof(i32));
#ifdef WORDS_BIGENDIAN
      // convert from little endian
      i32 = GINT32_FROM_LE(i32);
#endif
      frameOffs.push_back(i32);
      i++;
    }
	}

	// compare file date and date in project file -- write
	
	return true;
}

void MP3block::scanOffs() {
  // read frame legths
  guint32 MPEGhdr;

	file.seekg(0, ios::end);        // get file size	
	int file_size = file.tellg();

  file.seekg(beginMax, ios::beg);    // read offsets from whole file
  frameOffs.erase(frameOffs.begin(), frameOffs.end());    // clear offsets

	while (file.tellg() < endMax) {
    file.read(&MPEGhdr, sizeof(MPEGhdr));

    guint32 tmp = 0;
    tmp |= (MPEGhdr & 0x000000ff) << 24;
    tmp |= (MPEGhdr & 0x0000ff00) << 8;
    tmp |= (MPEGhdr & 0xff000000) >> 24;
    tmp |= (MPEGhdr & 0x00ff0000) >> 8;
    MPEGhdr = tmp;

  	if (!hdr_sync(MPEGhdr)) return;
  	
  	int size = hdr_frame_size(MPEGhdr);   // calculate frame length
  	int br = hdr_bitrate(MPEGhdr);
  	
  	if (br != bitrate) {
  	  vbr = true;
  	  if (br > max_vbr) max_vbr = br;
  	  if (br < bitrate) bitrate = br;
  	}
  	
  	if (file.tellg() + size - 4 >= file_size) {  // check if frame is not truncated
  	  endMax = file.tellg() - 4;                 // if yes then change block end
  	  end = endMax;
  	
#ifdef DEBUG_OUTPUT
    	cerr << "Trunc, Frames  " << frameOffs.size() << endl;
    	if (vbr) cerr << "VBR:  min: " << bitrate << "   max: " << max_vbr << endl;
#endif
  	  return;
  	}
  	
	  frameOffs.push_back(size);
	  file.seekg(file.tellg() + size - 4);      // set position to new frame start
	}
	
#ifdef DEBUG_OUTPUT
	cerr << "Frames  " << frameOffs.size() << endl;
 	if (vbr) cerr << "VBR:  min: " << bitrate << "   max: " << max_vbr << endl;
#endif
}

int MP3block::get_begin_sample_offset() {
  return get_begin_frame() * samples_per_frame();
}

int MP3block::get_end_sample_offset() {
  return (get_end_frame() + 1) * samples_per_frame();
}

void MP3block::set_begin_sample_offset(int v) {
  int offs = frameOffs.size();
  int frOffset = v / samples_per_frame();
  int sum = 0, index = 0;

  if (frOffset >= offs) {
    frOffset = offs;
    end = endMax;
  }
  else {
    while (index < frOffset) sum += frameOffs[index++];
    begin = sum;
  }

#ifdef DEBUG_OUTPUT
  cerr << "total frames " << offs << endl;
  cerr << "new begin " << begin << "   frame " << index << endl;
#endif

  tot_samples = (get_end_frame() - get_begin_frame()) * samples_per_frame();
  block_changed.emit();
}

void MP3block::set_end_sample_offset(int v) {
  int offs = frameOffs.size();
  int frOffset = v / samples_per_frame();
  int sum = 0, index = 0;

  if (frOffset >= offs) {
    frOffset = offs;
    end = endMax;
  }
  else {
    while (index <= frOffset) sum += frameOffs[index++];
    end = sum;
  }

#ifdef DEBUG_OUTPUT
  cerr << "new end " << end << "   frame " << index << endl;
#endif

  tot_samples = (get_end_frame() - get_begin_frame()) * samples_per_frame();
  block_changed.emit();
}

int MP3block::samples_per_frame() {
   if (layer == 1)
    	return 384;                 // layer 1:     384 samples per frame
    else
      return 1152;                // layer 2, 3: 1152 samples per frame
}

int MP3block::get_begin_frame() {
  int i = 0;
  int tmp = 0;
  int tmpsize = static_cast<int>(frameOffs.size());

  while (i < tmpsize) {
    tmp += frameOffs[i];
    if (tmp >= begin) break;
    i++;
  }

  return i;
}

int MP3block::get_end_frame() {
  int i = 0;
  int tmp = 0;
  int tmpsize = static_cast<int>(frameOffs.size());

  while (i < tmpsize) {
    tmp += frameOffs[i];
    if (tmp >= end) break;
    i++;
  }

  return i;
}

void MP3block::scanning_canceled() {
	if (scanning) {
    scanning = false;

    progress->destroy();

    scannedData.insert(scannedData.end(), num - scannedData.size(), static_cast<short>(32767));
    scan_done.emit();
  }
}


void MP3block::export_WAV(ofstream &ostr, int chan, int f, int bps, DialogProgress *dp,
                         string decoder_path, string conversion_path) {

  int pos = fileName.rfind(G_DIR_SEPARATOR);

  if (pos < 0)
    pos = 0;
  else
    pos++;

  exporting_WAV = true;

  string base_name(fileName, pos, fileName.size());
  dp->set_label(_("Exporting file: ") + base_name);

	if (!assigned) {
#ifdef DEBUG_OUTPUT
    cerr << "MP3block export: not assigned" << endl;
#endif
    return;
  }

  exporting = true;
  block_num = 0;
	export_progress = dp;

	audiobuffer = new char[audiobuffer_size];	

	if (!file.is_open()) file.open(fileName.c_str(), ios::binary | ios::in);			// open file

  file.seekg(begin, ios::beg);

  vector<string> s;

  if (chan != channels) {
    if (chan == 1) s.push_back("--mono");
    else s.push_back("--stereo");
  }

  if (bps == 8) s.push_back("--8bit");

  if (f != freq) {
    strstream tmp;
    tmp << f;

    string fstr;
    tmp >> fstr;

    s.push_back("-r");
    s.push_back(fstr);
  }

  s.push_back("-s");
  s.push_back("-");
  proc.start(decoder_path, s);

	export_stream = &ostr;
	
	play_to_ms = -1;
	play_from_ms = -1;
	data_written();
}


void MP3block::export_MP3(ofstream &ostr, string encoder_name, string encoder_path, int export_bitrate, int export_frequency,
		                     MP3block::ChannelMode channel_mode, bool export_crc, bool export_priv, bool export_original,
		                     bool export_copyright, vector<string> VBR_options, vector<string> encoder_options,
		                     bool always_process, bool strict_equal, DialogProgress *dp) {
		
  int pos = fileName.rfind(G_DIR_SEPARATOR);

  if (pos < 0)
    pos = 0;
  else
    pos++;

  exporting_WAV = false;  // exporting MP3 not WAV

  string base_name(fileName, pos, fileName.size());
  dp->set_label(_("Exporting file: ") + base_name);

	if (!assigned) {
#ifdef DEBUG_OUTPUT
    cerr << "MP3block export: not assigned" << endl;
#endif
    return;
  }

	play_to_ms = -1;
	play_from_ms = -1;

  playedMS = 0;
	play_from_offs_ms = 0;
	totalReceived = 0;
	usedBuffer = 0;
  sizeofpausebuffer = 0;
	
  // compare export and block parameters
  bool equal;

  if (always_process)
    equal = false;
  else
    if (strict_equal) {
      equal = export_bitrate == bitrate && export_frequency == freq && channel_mode == stereo_mode
              && export_crc == crc && export_priv == private_flag && export_original == original
              && export_copyright == copyright;
    }
    // strict equal is not required
    else {
      equal = export_bitrate == bitrate && export_frequency == freq && channel_mode == stereo_mode
              && export_crc == crc;
    }

#ifdef DEBUG_OUTPUT
    cerr << "MP3 export: strict equal = " << strict_equal << endl;
    cerr << "MP3 export:        equal = " << equal << endl;
#endif

  if (equal) {
#ifdef DEBUG_OUTPUT
    cerr << "MP3 export: copying data ..." << endl;
#endif
    // export parameters are equal to block param., only copy data
    int bsize = 4096;
    char *buf = new char[bsize];

    if (!file.is_open()) file.open(fileName.c_str(), ios::binary | ios::in);			// open file
    file.seekg(begin, ios::beg);

    int n = (end - begin) / bsize;
    int rem = (end - begin) % bsize;
    int total = n;

    while (n > 0) {
      file.read(buf, bsize);
      ostr.write(buf, bsize);

      // refresh progress bar
      if (n % 20 == 0) {
        dp->set_progress(100.0f - (static_cast<float>(n) / total * 100));
     	  while (Gtk::Main::events_pending() > 0) Gtk::Main::iteration();
     	
     	  // check whether export was canceled
     	  if (dp->cancel_state()) {
     	    delete[] buf;
          export_done.emit();
          return;
     	  }
      }

      n--;
    }

    if (rem > 0) {
      file.read(buf, rem);
      ostr.write(buf, rem);
      dp->set_progress(100.0f);
    }

    delete[] buf;
    export_done.emit();
  }
  // MP3 files are not equal, conversion is required
  else {
#ifdef DEBUG_OUTPUT
    cerr << "MP3 export: converting data ... " << encoder_path << endl;
#endif
    exporting = true;
    block_num = 0;
  	export_progress = dp;
  	export_stream = &ostr;
  	audiobuffer = new char[audiobuffer_size];	

  	if (!file.is_open()) file.open(fileName.c_str(), ios::binary | ios::in);			// open file
    file.seekg(begin, ios::beg);

    // encoder command line parameters
    vector<string> params;

    // insert encoder options
    params.insert(params.begin(), encoder_options.begin(), encoder_options.end());

    if (encoder_name == "lame") {
      if (export_bitrate > 0) {     // CBR
        params.push_back("-b");
        strstream ss;
        string s;

        ss << export_bitrate;
        ss >> s;
        params.push_back(s);
      }
      else {
        // set VBR option
        params.push_back("-v");
        // insert VBR options
        params.insert(params.end(), VBR_options.begin(), VBR_options.end());
      }

      // CRC option
      if (export_crc) params.push_back("-p");
      // original option
      if (!export_original) params.push_back("-o");
      // copyright option
      if (export_copyright) params.push_back("-c");
      // lame doesn't allow to set private flag

      // set channel mode parameter
      params.push_back("-m");
      switch(channel_mode) {
        case(MODE_STEREO) :
        case(MODE_DUAL)        : params.push_back("s"); break;    // dual flag not supported, set stereo
        case(MODE_JOINT_STEREO): params.push_back("j"); break;
        case(MODE_MONO)        : params.push_back("m"); break;
      }

      // input is MP3
      params.push_back("--mp3input");

      // set output frequency
      params.push_back("--resample");
      strstream strstr, strstr2;
      string str1;
      string str2;
      strstr << export_frequency / 1000;
      strstr >> str1;
      strstr2.width(3);
      strstr2.fill('0');
      strstr2 << export_frequency % 1000;
      strstr2 >> str2;
      params.push_back(str1 + "." + str2);

      // set stdin input
      params.push_back("-");
      // set stdout output
      params.push_back("-");
    }

    if (!file.is_open()) file.open(fileName.c_str(), ios::binary | ios::in);			// open file
    file.seekg(begin, ios::beg);
    audiobuffer = new char[audiobuffer_size];	

    proc.start(encoder_path, params);
    data_written();
  }
}

