/***************************************************************************
              ossdevice.cpp  -  Implementation of OSS interface
             ---------------------------------------------------
    begin                : Tue Feb 1 2000
    copyright            : (C) 2000 by Ladislav Slezk
    email                : slezak@penguin.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <iostream.h>

#include <gnome--/main.h>
#include <libgnome/gnome-i18n.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <unistd.h>						// close
#include <time.h>	

#include "ossdevice.h"

OSSdevice::OSSdevice() {
  OSSopened = false;
}

OSSdevice::~OSSdevice() {
  if (OSSopened) close_device();
}

bool OSSdevice::open_device() {
  if ((audio_fd = open(device.c_str(), O_WRONLY)) == -1) {
    cerr << _("Error opening sound device: ") << device << "  error: " << sys_errlist[errno] << endl;		// open failed
    return false;
  }

  OSSopened = true;

  int arg = 0x0008000C;			// 0008 = 8 fragments, 000C = 4kB each fragment; buffer = 32kB

  if (ioctl(audio_fd, SNDCTL_DSP_SETFRAGMENT, &arg) != 0)
  	cerr << _("Warning: Can not set buffer size!") << endl;

  audio_buf_info info;
  ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &info);
  fragment_size = info.fragsize;
	bufferSize = info.bytes;

  set_format(bit8);	// set default settings: 8bit, mono, 8kHz
  set_number_of_channels(1);
  set_frequency(8000);		// frequency should be last setting

  return true;
}

bool OSSdevice::opened() {
  return OSSopened;
}

bool OSSdevice::close_device() {

  if (OSSopened) {
	  audio_buf_info info;
    ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &info);

    while (info.bytes != bufferSize) {
//       usleep(10000);                      // sleep 100ms

  	  if (Gtk::Main::events_pending() > 0)
    	  Gtk::Main::iteration();
  	  else
        usleep(1000);											 // sleep 1ms

       ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &info);
    }

    if (close(audio_fd) == 0) {
      OSSopened = false;				// succes
      freq = 0;
      channels = 0;
      return true;
    }
    else
      return false;
  }
  return false;
}

bool OSSdevice::set_frequency(int f) {
  int speed = f;

  if (OSSopened) {
    if (ioctl(audio_fd, SNDCTL_DSP_SPEED, &speed) == -1)  {
      cerr << _("Error setting frequency: ") << sys_errlist[errno] << endl;
      return false;
	  }
		freq = speed;
	  return true;
  }
  else
    return false;
}

int OSSdevice::get_frequency() {
	return freq;
}

bool OSSdevice::set_number_of_channels(int ch) {
  int stereo = ch - 1;

  if (OSSopened) {

    if (ioctl(audio_fd, SNDCTL_DSP_STEREO, &stereo) == -1) {
      cerr << _("Error setting channels: ") << sys_errlist[errno] << endl;
      return false;
	  }
		channels = stereo;
	  return true;
  }
  else
    return false;
}

int OSSdevice::get_number_of_channels() {
	return channels;
}

bool OSSdevice::play(char *data, int len) {
  audio_buf_info info;

  if (!OSSopened) return false;

	if (len <= bufferSize) {		// do non-blocking write if possible -- CPU load is lesser
    ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &info);

    while (info.bytes < len) {
      usleep(1000);                                  // sleep 1ms

////////////////////////////////////////
//  	  if (Gtk::Main::events_pending() > 0)
//  	    Gtk::Main::iteration();
//  	  else
//        usleep(1);
////////////////////////////////////////

      ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &info);
    }
  }

  if (write(audio_fd, data, len) == -1) {
    cerr << _("Error writing to sound device: ") << sys_errlist[errno] << endl;
    return false;
  }

  return true;
}

bool OSSdevice::stop() {
  if (!OSSopened) return false;

  if (ioctl(audio_fd, SNDCTL_DSP_RESET, 0) == -1) return false;
  return true;
}

string OSSdevice::get_info() {
	return string(_("Open Sound System output"));
}

bool OSSdevice::set_format(bits b) {
  int formatx;

  format = b;
  switch (b) {
    case (bit8)     : formatx = AFMT_U8; break;       // audio format: unsigned 8 bit
    case (bit16_le) :	formatx = AFMT_S16_LE; break;   //  signed 16 bit little endian
    case (bit16_be) :	formatx = AFMT_S16_BE; break;   //  signed 16 bit big endian
  }

	if (ioctl(audio_fd, SNDCTL_DSP_SETFMT, &formatx) == -1) { /* Fatal error */
   	cerr << _("Error setting data format: ") << sys_errlist[errno] << endl;
   	return false;
  }

  return true;
}

SoundDevice::bits OSSdevice::get_format() {
  return format;
}

void OSSdevice::reset(){
  if (OSSopened) ioctl(audio_fd, SNDCTL_DSP_RESET, 0);
}

int OSSdevice::get_processed_bytes() {
  if (OSSopened) {
    count_info info;

    ioctl(audio_fd, SNDCTL_DSP_GETOPTR, &info);
    return info.bytes;
  }
  else
    return 0;
}

int OSSdevice::get_processed_ms() {
  int bps = format + 1;
  int ch = channels + 1;

  return static_cast<int>(1000.0 * get_processed_bytes() / bps / ch / freq);
}
