/***************************************************************************
            subprocess.cpp  -  execute command on background
           --------------------------------------------------
    begin                : Tue Oct 3 2000
    copyright            : (C) 2000 by Ladislav Slezk
    email                : slezak@penguin.cz
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "subprocess.h"

#include <signal.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <string.h>

#include <gnome--/app.h>
#include <libgnome/gnome-i18n.h>

// initialize static members
bool SubProcess::handler_installed = false;
map<pid_t, SubProcess*> SubProcess::pidmap;

SubProcess::SubProcess() : buffer_size(4096) {
  pid = 0;
  run = false;
  buffer_in_size = 0;
  close_in = false;

  // install SIGCHLD signal handler
  if (!handler_installed) {
#ifdef DEBUG_OUTPUT
    cerr << "Installing signal handler ..." << endl;
#endif
    struct sigaction act;
    act.sa_handler = signal_handler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = SA_NOCLDSTOP;    // ignore child stop

    if (sigaction(SIGCHLD, &act, NULL) < 0) {
      cerr << _("Error setting SIGCHLD handler!!") << endl;
      exit(1);
    }

    // set standard signal handler (coredump) instead of displaying dialog
    act.sa_handler = SIG_DFL;
    if (sigaction(SIGSEGV, &act, (struct sigaction *)NULL) < 0) {
      cerr << "Error setting SIGSEGV handler!!" << endl;
      exit(1);
    }

    handler_installed = true;
  }

  // allocate buffers
  buffer_out = new char[buffer_size];                 // allocate buffer
  buffer_err = new char[buffer_size];                 // allocate buffer
  buffer_in = 0;
}

SubProcess::~SubProcess() {
  c_err.disconnect();
  c_out.disconnect();
  c_in.disconnect();

  if (run) kill_process();
  if (buffer_in != 0) delete[] buffer_in;
  delete[] buffer_out;
  delete[] buffer_err;
}

bool SubProcess::start(string command, const vector<string> &args) {
  if (run) return false;        // subprocess is already running

  // create pipes
  if (pipe(stdout_pipe)) return false;
  if (pipe(stderr_pipe)) return false;
  if (pipe(stdin_pipe)) return false;

  pid = fork();     // fork process

  if (pid < 0) {    // fork() failed
    return false;
  }

  if (pid == 0) {   // child process
    // assign std{out, err, in} to pipe
    dup2(stdout_pipe[1], STDOUT_FILENO);
    dup2(stderr_pipe[1], STDERR_FILENO);
    dup2(stdin_pipe[0], STDIN_FILENO);

    // close all pipes
    close(stdout_pipe[0]);
    close(stdout_pipe[1]);
    close(stderr_pipe[0]);
    close(stderr_pipe[1]);
    close(stdin_pipe[0]);
    close(stdin_pipe[1]);

    int num = args.size();
    int index = 1;

    // create array of arguments
    char **arg = new char*[num + 2];                  // number of parameters + command + terminating NULL
    arg[0] = const_cast<char*>(command.c_str());      // first parameter is command

    int id = 0;
    while (num > 0) {
      char *s = new char[args[id].size() + 1];
      strcpy(s, args[id].c_str());
      arg[index++] = s;
      num--;
      id++;
    }

    arg[index] = NULL;                    // terminate parameter list

#ifdef DEBUG_OUTPUT
    cerr << "Subprocess parameters: ";
    copy(args.begin(), args.end(), ostream_iterator<string>(cerr, " "));
    cerr << endl;
#endif

    execvp(command.c_str(), arg);           // execute command
    cerr << ("execvp(): Error executing command: ") << command.c_str() << endl;    // exec failed
    exit(-1);
  }

  // parent process

  // close ends for writing
  close(stdout_pipe[1]);
  close(stderr_pipe[1]);

  // close end for reading
  close(stdin_pipe[0]);

  // reset data sizes
  std_out_len = 0;
  std_err_len = 0;
  std_in_len = 0;

  // set non-blocking operations
  int flag = fcntl(stdout_pipe[0], F_GETFL, 0);   // read flag
  flag = flag | O_NONBLOCK;                       // add non-block attribut
  fcntl(stdout_pipe[0], F_SETFL, flag);           // set flag

  flag = fcntl(stderr_pipe[0],F_GETFL,0);
  flag = flag | O_NONBLOCK;
  fcntl(stderr_pipe[0], F_SETFL, flag);

  flag = fcntl(stdin_pipe[1],F_GETFL,0);
  flag = flag | O_NONBLOCK;
  fcntl(stdin_pipe[1], F_SETFL, flag);

  // watch stdout for data
  c_out = Gtk::Main::input.connect(slot(this, &SubProcess::canReadStdOut),
                                   stdout_pipe[0], GDK_INPUT_READ);

  // watch stderr for data
  c_err = Gtk::Main::input.connect(slot(this, &SubProcess::canReadStdErr),
                                   stderr_pipe[0], GDK_INPUT_READ);

  run = true;
  close_in = false;

  pidmap[pid] = this;           // add pid and SubProcess to pidmap
  return true;
}

bool SubProcess::kill_process(int sig = SIGTERM) {
  c_err.disconnect();
  c_out.disconnect();
  c_in.disconnect();

  close(stdin_pipe[1]);
  close(stderr_pipe[0]);
  close(stdout_pipe[0]);

  if (pid > 0)
    return kill(pid, sig);
  else
    return false;
}


void SubProcess::close_stdin() {
  if (std_in_len > 0 && run) {
    // some data are not written yet
    close_in = true;
  }
  else
    close(stdin_pipe[1]);
}


void SubProcess::write_stdin(const char *buff, int size) {
  if (size <= 0) return;     // buffer is not empty

  if (buffer_in_size < size) {
    if (buffer_in != 0) delete buffer_in;
    buffer_in = new char[size];             // reallocate buffer
    buffer_in_size = size;
  }

  // copy data to buffer
  memcpy(buffer_in, buff, size);

  std_in_len = size;
  pos_in = buffer_in;

  // watch stdin for writing
  c_in = Gtk::Main::input.connect(slot(this, &SubProcess::canWriteStdIn),
                                     stdin_pipe[1], GDK_INPUT_WRITE);
}

void SubProcess::canWriteStdIn(int, GdkInputCondition) {
  int writed = write(stdin_pipe[1], pos_in, std_in_len);

  if (writed != std_in_len) {
    pos_in += writed;              // not all writed
    std_in_len -= writed;
  }
  else {
    c_in.disconnect();
    data_written.emit();
    if (close_in) close(stdin_pipe[1]);
  }
//  cerr << "Subprocess: writed: " << writed << endl;
}


void SubProcess::canReadStdOut(int, GdkInputCondition) {
  std_out_len = read(stdout_pipe[0], buffer_out, buffer_size);

  if (std_out_len > 0) {
    stdout_data.emit(buffer_out, std_out_len);
  }

  if (std_out_len == 0) {
    c_out.disconnect();         // process exited - disconnect
  }

}

void SubProcess::canReadStdErr(int, GdkInputCondition) {
  std_err_len = read(stderr_pipe[0], buffer_err, buffer_size);

  if (std_err_len > 0) {
    stderr_data.emit(buffer_err, std_err_len);
  }

  if (std_err_len == 0) {
    c_err.disconnect();
  }
}

void SubProcess::signal_handler(int s) {
  // this is signal handler -- calling GUI function is not allowed!!
  // tell Gtk to call get_status()
  Gtk::Main::idle.connect(SigC::slot(&get_status));
}

void SubProcess::exited(int stat) {
  process_exited.emit(stat);
  run = false;
}

int SubProcess::get_status() {
  int status;
  SubProcess *sp;

  pid_t pid_exited = waitpid(-1, &status, WNOHANG);     // test all child processes

  while (pid_exited > 0) {

#ifdef DEBUG_OUTPUT
    cerr << "Subprocess: Process exited : " << pid_exited << endl;
#endif

    // select SubProcess object which handles process pid_exited
    sp = pidmap[pid_exited];

    if (sp != 0)
      sp->exited(WEXITSTATUS(status));
    else
      cerr << "Subprocess: Unknown process exited!" << endl;

    pidmap.erase(pid_exited);     // erase exited pid from pidmap

    pid_exited = waitpid(-1, &status, WNOHANG);
  }

  return false;       // do not recall again this idle function
}

